if (!window.c2cgeoform) {
  c2cgeoform = {};
}

// An object including oids of search fields that have already
// been initialized. This is required because of jQuery steps.
c2cgeoform.searchfields = {};
